@@ -16,20 +16,110 @@
 //! Convenience data structures built from the IPC structs.
 
 use crate::config as cfg;
-use crate::ipc;
-use crate::ipc::NodeMethods;
 use crate::util;
 use crate::util::DisplayFormat;
 use lazy_static::lazy_static;
+use serde::{Deserialize, Serialize};
 use std::cmp;
 use std::collections::HashMap;
 use swayipc as s;
 
+/// Immutable Node Iterator
+///
+/// Iterates nodes in depth-first order, tiled nodes before floating nodes.
+pub struct NodeIter<'a> {
+    stack: Vec<&'a s::Node>,
+}
+
+impl<'a> NodeIter<'a> {
+    pub fn new(node: &'a s::Node) -> NodeIter {
+        NodeIter { stack: vec![node] }
+    }
+}
+
+impl<'a> Iterator for NodeIter<'a> {
+    type Item = &'a s::Node;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        if let Some(node) = self.stack.pop() {
+            for n in &node.floating_nodes {
+                self.stack.push(&n);
+            }
+            for n in &node.nodes {
+                self.stack.push(&n);
+            }
+            Some(node)
+        } else {
+            None
+        }
+    }
+}
+
+/// Extension methods for [`swayipc::Node`].
+pub trait NodeMethods {
+    /// Returns an iterator for this [`swayipc::Node`] and its childres.
+    fn iter(&self) -> NodeIter;
+
+    fn is_window(&self) -> bool;
+
+    /// Either a workspace or a con holding windows, e.g. a vertical split side
+    /// in a horizontally split workspace.
+    fn is_container(&self) -> bool;
+
+    /// Returns all nodes being application windows.
+    fn windows(&self) -> Vec<&s::Node>;
+
+    /// Returns all nodes being workspaces.
+    fn workspaces(&self) -> Vec<&s::Node>;
+
+    fn is_scratchpad(&self) -> bool;
+}
+
+impl NodeMethods for s::Node {
+    fn iter(&self) -> NodeIter {
+        NodeIter::new(self)
+    }
+
+    fn is_window(&self) -> bool {
+        (self.node_type == s::NodeType::Con
+            || self.node_type == s::NodeType::FloatingCon)
+            && self.name.is_some()
+    }
+
+    fn is_container(&self) -> bool {
+        self.node_type == s::NodeType::Workspace
+            || self.node_type == s::NodeType::Con
+                && self.name.is_none()
+                && self.layout != s::NodeLayout::None
+    }
+
+    fn windows(&self) -> Vec<&s::Node> {
+        self.iter().filter(|n| n.is_window()).collect()
+    }
+
+    fn workspaces(&self) -> Vec<&s::Node> {
+        self.iter()
+            .filter(|n| n.node_type == s::NodeType::Workspace)
+            .collect()
+    }
+
+    fn is_scratchpad(&self) -> bool {
+        self.name.is_some() && self.name.as_ref().unwrap().eq("__i3_scratch")
+    }
+}
+
+/// Extra properties gathered by swayrd for windows and workspaces.
+#[derive(Clone, Debug, Deserialize, Serialize)]
+pub struct ExtraProps {
+    /// Milliseconds since UNIX epoch.
+    pub last_focus_time: u128,
+}
+
 #[derive(Debug)]
 pub struct Window<'a> {
     node: &'a s::Node,
     workspace: &'a s::Node,
-    extra_props: Option<ipc::ExtraProps>,
+    extra_props: Option<ExtraProps>,
 }
 
 impl Window<'_> {
@@ -108,39 +198,36 @@ lazy_static! {
         regex::Regex::new("(.+)(-[0-9.]+)").unwrap();
 }
 
+fn maybe_html_escape(do_it: bool, text: &str) -> String {
+    if do_it {
+        text.replace("<", "&lt;")
+            .replace(">", "&gt;")
+            .replace("&", "&amp;")
+    } else {
+        text.to_string()
+    }
+}
+
 impl<'a> DisplayFormat for Window<'a> {
     fn format_for_display(&self, cfg: &cfg::Config) -> String {
-        let default_format = cfg::Format::default();
-        let fmt = cfg
-            .format
-            .as_ref()
-            .and_then(|f| f.window_format.as_ref())
-            .unwrap_or_else(|| default_format.window_format.as_ref().unwrap());
-        let urgency_start = cfg
-            .format
-            .as_ref()
-            .and_then(|f| f.urgency_start.as_ref())
-            .unwrap_or_else(|| default_format.urgency_start.as_ref().unwrap());
-        let urgency_end = cfg
-            .format
-            .as_ref()
-            .and_then(|f| f.urgency_end.as_ref())
-            .unwrap_or_else(|| default_format.urgency_end.as_ref().unwrap());
-        let icon_dirs = cfg
-            .format
-            .as_ref()
-            .and_then(|f| f.icon_dirs.as_ref())
-            .unwrap_or_else(|| default_format.icon_dirs.as_ref().unwrap());
+        let window_format = cfg.get_format_window_format();
+        let urgency_start = cfg.get_format_urgency_start();
+        let urgency_end = cfg.get_format_urgency_end();
+        let html_escape = cfg.get_format_html_escape();
+        let icon_dirs = cfg.get_format_icon_dirs();
         // fallback_icon has no default value.
-        let fallback_icon: Option<&String> =
-            cfg.format.as_ref().and_then(|f| f.fallback_icon.as_ref());
+        let fallback_icon: Option<Box<std::path::Path>> = cfg
+            .get_format_fallback_icon()
+            .as_ref()
+            .map(|i| std::path::Path::new(i).to_owned().into_boxed_path());
 
         // Some apps report, e.g., Gimp-2.10 but the icon is still named
         // gimp.png.
         let app_name_no_version =
             APP_NAME_AND_VERSION_RX.replace(self.get_app_name(), "$1");
 
-        fmt.replace("{id}", format!("{}", self.get_id()).as_str())
+        window_format
+            .replace("{id}", format!("{}", self.get_id()).as_str())
             .replace(
                 "{urgency_start}",
                 if self.is_urgent() {
@@ -157,33 +244,49 @@ impl<'a> DisplayFormat for Window<'a> {
                     ""
                 },
             )
-            .replace("{app_name}", self.get_app_name())
+            .replace(
+                "{app_name}",
+                &maybe_html_escape(html_escape, self.get_app_name()),
+            )
             .replace(
                 "{workspace_name}",
-                self.workspace.name.as_ref().unwrap().as_str(),
+                &maybe_html_escape(
+                    html_escape,
+                    self.workspace.name.as_ref().unwrap().as_str(),
+                ),
+            )
+            .replace(
+                "{marks}",
+                &maybe_html_escape(html_escape, &self.node.marks.join(", ")),
             )
             .replace(
                 "{app_icon}",
-                util::get_icon(self.get_app_name(), icon_dirs)
-                    .or_else(|| util::get_icon(&app_name_no_version, icon_dirs))
+                util::get_icon(self.get_app_name(), &icon_dirs)
+                    .or_else(|| {
+                        util::get_icon(&app_name_no_version, &icon_dirs)
+                    })
                     .or_else(|| {
                         util::get_icon(
                             &app_name_no_version.to_lowercase(),
-                            icon_dirs,
+                            &icon_dirs,
                         )
                     })
-                    .or_else(|| fallback_icon.map(|f| f.to_string()))
+                    .or(fallback_icon)
+                    .map(|i| i.to_string_lossy().into_owned())
                     .unwrap_or_else(String::new)
                     .as_str(),
             )
-            .replace("{title}", self.get_title())
+            .replace(
+                "{title}",
+                &maybe_html_escape(html_escape, self.get_title()),
+            )
     }
 }
 
 fn build_windows<'a>(
     root: &'a s::Node,
     include_scratchpad_windows: bool,
-    extra_props: Option<&HashMap<i64, ipc::ExtraProps>>,
+    extra_props: Option<&HashMap<i64, ExtraProps>>,
 ) -> Vec<Window<'a>> {
     let mut v = vec![];
     for workspace in root.workspaces() {
@@ -205,7 +308,7 @@ fn build_windows<'a>(
 fn build_workspaces<'a>(
     root: &'a s::Node,
     include_scratchpad: bool,
-    extra_props: Option<&HashMap<i64, ipc::ExtraProps>>,
+    extra_props: Option<&HashMap<i64, ExtraProps>>,
 ) -> Vec<Workspace<'a>> {
     let mut v = vec![];
     for workspace in root.workspaces() {
@@ -240,7 +343,7 @@ fn build_workspaces<'a>(
 pub fn get_windows<'a>(
     root: &'a s::Node,
     include_scratchpad_windows: bool,
-    extra_props: Option<&HashMap<i64, ipc::ExtraProps>>,
+    extra_props: Option<&HashMap<i64, ExtraProps>>,
 ) -> Vec<Window<'a>> {
     let extra_props_given = extra_props.is_some();
     let mut wins = build_windows(root, include_scratchpad_windows, extra_props);
@@ -254,7 +357,7 @@ pub fn get_windows<'a>(
 pub fn get_workspaces<'a>(
     root: &'a s::Node,
     include_scratchpad: bool,
-    extra_props: Option<&HashMap<i64, ipc::ExtraProps>>,
+    extra_props: Option<&HashMap<i64, ExtraProps>>,
 ) -> Vec<Workspace<'a>> {
     let mut workspaces =
         build_workspaces(root, include_scratchpad, extra_props);
@@ -314,7 +417,7 @@ pub fn select_workspace_or_window<'a>(
 
 pub struct Workspace<'a> {
     node: &'a s::Node,
-    extra_props: Option<ipc::ExtraProps>,
+    extra_props: Option<ExtraProps>,
     pub windows: Vec<Window<'a>>,
 }
 
@@ -364,16 +467,10 @@ impl PartialOrd for Workspace<'_> {
 
 impl<'a> DisplayFormat for Workspace<'a> {
     fn format_for_display(&self, cfg: &cfg::Config) -> String {
-        let default_format = cfg::Format::default();
-        let fmt = cfg
-            .format
-            .as_ref()
-            .and_then(|f| f.workspace_format.as_ref())
-            .unwrap_or_else(|| {
-                default_format.workspace_format.as_ref().unwrap()
-            });
+        let fmt = cfg.get_format_workspace_format();
+        let html_escape = cfg.get_format_html_escape();
 
         fmt.replace("{id}", format!("{}", self.get_id()).as_str())
-            .replace("{name}", self.get_name())
+            .replace("{name}", &maybe_html_escape(html_escape, self.get_name()))
     }
 }
